/* * * * * * * * * * * *\
 *    Panel-Docklet    *
 *    ~~~~~~~~~~~~~    *
 *        Jodli        *
 *  davidlewe@gmail.at *
\* * * * * * * * * * * */


const Main = imports.ui.main;
const Shell = imports.gi.Shell;
const Lang = imports.lang;
const St = imports.gi.St;
const Gtk = imports.gi.Gtk;
const Mainloop = imports.mainloop;
const LayoutManager = Main.layoutManager;
const Tweener = imports.ui.tweener;
const DND = imports.ui.dnd;
const Clutter = imports.gi.Clutter;
const Pango = imports.gi.Pango;
const Meta = imports.gi.Meta;
const PopupMenu = imports.ui.popupMenu;
const AppFavorites = imports.ui.appFavorites;
const MessageTray = imports.ui.messageTray;

const Extension = imports.misc.extensionUtils.getCurrentExtension();
const Convenience = Extension.imports.convenience;
const Gettext = imports.gettext;
const _ = Gettext.domain("panel-docklet").gettext;
const Tracker = Shell.WindowTracker.get_default();

const Gio = imports.gi.Gio;
const GLib = imports.gi.GLib;
const PanelDocklet_iface = '<node> \
<interface name="org.gnome.Shell.Extensions.panelDocklet"> \
<method name="createDock"> \
    <arg type="i" direction="in" name="i" /> \
</method> \
<method name="deleteDock"> \
    <arg type="i" direction="in" name="i" /> \
</method> \
<method name="selectDocklet"> \
    <arg type="i" direction="in" name="i" /> \
</method> \
<method name="deselectDocklet"> \
    <arg type="i" direction="in" name="i" /> \
</method> \
<method name="get_monitors"> \
    <arg type="aai" direction="out" /> \
</method> \
</interface> \
</node>';

//WARNING: padding-values also include border!
const DOCK_POS = {PANEL_LEFT:0, PANEL_MIDDLE_LEFT:1, PANEL_MIDDLE_RIGHT:2, PANEL_RIGHT:3, TOP:10, RIGHT:11, BOTTOM:12, LEFT:13},
	MIDDLE_CLICK = {CLOSE:0, MINIMIZE:1, NEW:2, QUIT: 3},
	WIN_TEXTURE_MAX_SIZE = 250,
	ICON_PADDING_SUM = 6, // -> from stylesheet
	DOCK_PADDING_SHORT_SUM = 1, // -> from stylesheet
	DOCK_PADDING_LONG_SUM = 12, // -> from stylesheet
	HOVER_MENU_LENTGH = 250,
	MIN_LABEL_LENGTH = 20,
	WORKSPACE_INDICATOR = {PADDING_SUM:8, BUTTON_PADDING_SUM: 0, BUTTON_MARGIN: 2}, //PADDING_SUM, BUTTON_PADDING_SUM -> from stylesheet
	MIN_OVERHANG_DIFF = 15, // value also has to consider DOCK_PADDING_SHORT_SUM
	MIN_ICON_SIZE = 12; //has to be bigger than ICON_PADDING_SUM

const FAV_WIN_MASC = {
	window: {
		is_on_all_workspaces: function() {return false;},
		has_focus: function() {return false;}
	}
};
let Settings;


let get_arrow = (PopupMenu.unicodeArrow) ? PopupMenu.unicodeArrow : PopupMenu.arrowIcon; //arrowIcon: gnome 3.16


const RightClickMenu = new Lang.Class({
	Name: "RightClickMenu",
	Extends: PopupMenu.PopupMenu,
	
	dock: null,
	
	_init: function(dock, arrowSide) {
        this.parent(dock.actor, 0.5, arrowSide);
		this.dock = dock;
		this.actor.add_style_class_name("panelDocklet_popup");
        
        this.connect("menu-closed", Lang.bind(this, this.menu_closed));
	},
	
	menu_closed: function() {
		if(this.dock.settings.autohide)
			this.dock.hide();
	},
	_add_action_button: function(label, parent, fu) {
		let item = new St.Button({style_class: "action_button", label: label, reactive:true, can_focus:true});
		parent.add(item);
		item.connect("clicked", Lang.bind(this, fu));
		
		return item;
	},
	add_window_list: function(app, metaWin, pos) {
		let line = new PopupMenu.PopupBaseMenuItem({reactive:false, can_focus:false}),
			wins = app.get_windows(),
			//scroller = new St.ScrollView({width:400, height:100, vscrollbar_policy:Gtk.PolicyType.NEVER}),
			//scroller = new St.ScrollView({width:0, height:0}),
			scroller = new St.ScrollView(),
			win_items = new St.BoxLayout({vertical:this.dock.vertical}),
			currentWS = global.screen.get_active_workspace_index(),
			box, item, button, close_vertical, close, diff, width, height, win;
		
		//line.set_x_align(St.Align.END);
		//line.set_x_fill(false);
		for(let k in wins) {
			win = wins[k];
			if(win.is_skip_taskbar())
				continue;
			
			button = new St.Button({
					style_class: "win_texture",
					reactive:true,
					can_focus:true
				});
			button.connect("clicked", Lang.bind([win, this], function() {toggle_win(this[0]); this[1].itemActivated();}));
			
			if(win.has_focus())
				button.add_style_class_name("focused");
			else if(win.minimized)
				button.opacity = 128;
			if(win == metaWin)
				button.add_style_class_name("selected");
			
			box = new St.BoxLayout({vertical:true});
			button.set_child(box);
			
			
			box.add(new St.Label({text: win.title, width:WIN_TEXTURE_MAX_SIZE}));
			
			item = new Clutter.Clone({source: win.get_compositor_private().get_texture()}),
			[width, height] = item.get_size();
			
			diff = Math.abs(win.get_workspace().index() - currentWS)*50;
			
			if(width > height)
				item.set_size(WIN_TEXTURE_MAX_SIZE - diff, Math.round(height / (width / WIN_TEXTURE_MAX_SIZE)) - diff);
			else
				item.set_size(Math.round(width / (height / WIN_TEXTURE_MAX_SIZE)) - diff, WIN_TEXTURE_MAX_SIZE - diff);
			if(win.minimized)
				item.opacity = 128;
			box.add(item, {x_fill:false});
		
			win_items.add(button);
			
			close_vertical = new St.BoxLayout({height:0});
			box.add(close_vertical, {x_align:St.Align.END, x_fill:false});
			
			close = new St.Button({
				style_class: "notification-close",
				reactive:true,
				can_focus:true,
				translation_x: 10,
				translation_y: -item.get_height()
			});
			close_vertical.add(close);
			close.connect("clicked", Lang.bind([win, button, this], function() {
					this[0].delete(global.get_current_time());
					if(app.get_windows().length > 1) //window wont be closed before function is done
						this[1].destroy();
					else
						this[2].itemActivated();
				}));
		}
		
		
		
		scroller.add_actor(win_items);
		line.actor.add(scroller);
		this.addMenuItem(line, pos);
		
		if(this.dock.vertical) {
			scroller.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
			let monitor_h = LayoutManager.monitors[this.dock.settings.dock_monitor].height,
				win_box_h = win_items.get_transformed_size()[1],
				popup_h = this.box.get_transformed_size()[1];
			
			scroller.set_size(300, (monitor_h < popup_h) ? monitor_h - (popup_h - win_box_h)-70 : win_box_h);
		}
		else {
			scroller.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.NEVER);
			let monitor_w = LayoutManager.monitors[this.dock.settings.dock_monitor].width,
				win_box_w = win_items.get_transformed_size()[0],
				popup_w = this.box.get_transformed_size()[0];
			
			scroller.set_size((monitor_w < popup_w) ? monitor_w - (popup_w - win_box_w)-40 : win_box_w, 300);
		}
	},
	
	open_for_icon: function(icon) {
		this.removeAll();
		
		if(this.dock.hover_menu_isOpen)
			this.dock.close_hover_menu();
		
		let winBox = icon.winBox,
			app = icon.app,
			metaWin = winBox.window,
			line = new PopupMenu.PopupBaseMenuItem({reactive:false, can_focus:false}),
			action_items = new St.BoxLayout(),
			favs = AppFavorites.getAppFavorites(),
			item;
		line.actor.add(action_items, {expand: true});
		
		
		//Actions:
		if(!icon.is_fav_icon) {
			let monitor = metaWin.get_monitor(); //cant get that var outside: fav_icons dont have a get_monitor-function
			if(monitor) {
				item = this._add_action_button(null, action_items, function() {metaWin.move_to_monitor(monitor-1); this.itemActivated();});
				item.set_child(get_arrow(St.Side.LEFT));
			}
			if(monitor < LayoutManager.monitors.length-1) {
				item = this._add_action_button(null, action_items, function() {metaWin.move_to_monitor(monitor+1); this.itemActivated();});
				item.set_child(get_arrow(St.Side.RIGHT));
			}
			
			if(!metaWin.is_on_all_workspaces()) {
				item = this._add_action_button(null, action_items, function() {metaWin.change_workspace_by_index(metaWin.get_workspace().index()+1, true, true); this.itemActivated();});
				item.set_child(get_arrow(St.Side.BOTTOM));
			}
			if(metaWin.get_workspace().index() && !metaWin.is_on_all_workspaces()) {
				item = this._add_action_button(null, action_items, function() {metaWin.change_workspace_by_index(metaWin.get_workspace().index()-1, true, true); this.itemActivated();});
				item.set_child(get_arrow(St.Side.TOP));
			}
		}
		
		action_items.add(new St.Label( {text: app.get_name(), style_class:"action_label"}), {expand: true});
		
		
		if(!icon.is_fav_icon) {
			if(metaWin.minimized)
				this._add_action_button("\u21A5", action_items, function() {metaWin.unminimize(); metaWin.activate(global.get_current_time()); this.itemActivated();}); //arrow from bottom
			else
				this._add_action_button("_", action_items, function() {metaWin.minimize(); this.itemActivated();});
			if(metaWin.maximized_horizontally || metaWin.maximized_vertically)
				this._add_action_button("\u2B1C", action_items, function() { //square--
						metaWin.unmaximize(Meta.MaximizeFlags.VERTICAL);
						metaWin.unmaximize(Meta.MaximizeFlags.HORIZONTAL);
						metaWin.activate(global.get_current_time());
						this.itemActivated();
					});
			else
				//this._add_action_button("\u2610", action_items, function() { //square--
				this._add_action_button("\u21F1", action_items, function() { //arrow corner top
						metaWin.maximize(Meta.MaximizeFlags.VERTICAL);
						metaWin.maximize(Meta.MaximizeFlags.HORIZONTAL);
						metaWin.activate(global.get_current_time());
						this.itemActivated();
					});
			
			this._add_action_button("\u2715", action_items, function() {metaWin.delete(global.get_current_time()); this.itemActivated();});
		}
		this.addMenuItem(line);
		
		
		//window options:
		this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
		this.addAction(_("New window"), function() {app.open_new_window(global.screen.get_active_workspace_index());});
		if(!icon.is_fav_icon)
			this.addAction(_("Quit application"), function() {app.request_quit();});
		this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
		
		if(!icon.is_fav_icon) {
			if(metaWin.above)
				this.addAction(_("\u2714 Always on top"), function() {metaWin.unmake_above();});
			else
				this.addAction(_("\u25CB Always on top"), function() {metaWin.make_above();});
		
			if(metaWin.is_on_all_workspaces())
				this.addAction(_("\u2714 Always on visible workspace"), function() {metaWin.unstick();});
			else
				this.addAction(_("\u25CB Always on visible workspace"), function() {metaWin.stick();});
		}
		if(favs.isFavorite(app.get_id()))
			this.addAction(_("\u2714 In favorites"), function() {favs.removeFavorite(app.get_id());});
		else
			this.addAction(_("\u25CB In favorites"), function() {favs.addFavorite(app.get_id());});
		
		
		//icon-options:
		if(!icon.is_fav_icon) {
			this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
			this.addAction(_("Remove icon from docklet"), function() {while(icon.destroy(false, false, icon.winBox));}); //watch out: icon.winBox changes after a destroy()!
			if(icon.is_app_icon)
				this.addAction(_("Degroup application-icon"), function() {
						let dock = icon.dock,
							wins = icon.app_windows,
							i = wins.length,
							ws_index = icon.ws_index, //doesnt have to be the real ws_index because circumstances wont change
							win;
						delete dock.applications[ws_index][app.get_id()];
						
						while(i--) {
							win = wins[i];
							if(win == winBox)
								continue;
							dock.create_icon(win, ws_index, win.window.get_monitor(), false);
							icon.destroy(false, false, win);
						}
						icon.is_app_icon = false;
					});
			else
				this.addAction(_("Group all application-icons"), function() {
						let dock = icon.dock,
							ws_index = icon.ws_index,
							monitor = metaWin.get_monitor(),
							win = dock.windowList[ws_index],
							app_name = winBox._app.get_id();
						
						icon.set_app_icon(app_name);
						
						while(win) {
							if(win.app.get_id() == app_name && win != icon) {
								win.destroy(true, false, win.winBox); //its not an app-icon and its not the last icon to set as favorite - so destroy() can skip checking
								dock.create_icon(win.winBox, ws_index, monitor, true);
							}
							
							win = win.next;
						}
					
					});
			
			if(icon._label)
				this.addAction(_("Remove label"), function() {
						let dock = icon.dock;
						icon._label.destroy();
						icon._label = null;
						
						Tweener.addTween(icon.actor, obj_size({
								time: 0.4,
								transition: "easeOutQuint"
							}, dock.vertical, dock.icon_size));
						
						--dock.iconLabel_count;
						
						dock.update_icon_size();
					});
			else
				this.addAction(_("Add label"), function() {
						let dock = icon.dock;
						icon.create_label();
						Tweener.addTween(icon.actor, obj_size({
								time: 0.4,
								transition: "easeOutSine"
							}, dock.vertical, dock.icon_size + dock.label_length));
						
						++dock.iconLabel_count;
						dock.update_icon_size();
					});
			
		}
		
		
		//Windows:
		if(!icon.is_fav_icon) {
			if(icon.dock.settings.dock_orientation == DOCK_POS.BOTTOM) {
				this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
				this.add_window_list(app, metaWin);
			}
			else {
				this.add_window_list(app, metaWin, 2);
				this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem(), 3);
			}
		}
		this.sourceActor = icon.actor;
		this.open(true);
	}
});
const Docklet = new Lang.Class({
	Name: "Docklet",

	windowList: null, //list
	applications: null, //array
	fav_icons: null, //array
	settings: null, //array

	actor: null,
	windowActor: null,
	windowActor_vertical: null,
	hover_menu: null,
	
	id:0,
	_dock_x: 0,
	_dock_y: 0,
	_dock_hide_x: 0,
	_dock_hide_y: 0,
	currentWS: -1, //-1: to make sure no icons are added when indexing - there is a separate to_ws()-call for that so animations are smoother
	hover_menu_isOpen: false,
	vertical: false,
	intoPanel: false,
	hidden: false,
	selected: false,
	_hide_timeout: null,
	iconList_count: 0,
	iconLabel_count: 0,
	iconList_is_decreased: false,
	icon_space: 0, //if it becomes < 0, it wont have any effect, we dont have to care for that
	extra_space: DOCK_PADDING_LONG_SUM,

	_ID_overviewShown: null,
	_ID_overviewHidden: null,
	_ID_xdnd_drag_begin: null,
	_ID_xdnd_drag_end: null,
	
	_init: function(id) {
		this.windowList = [];
		this.applications = [];
		this.fav_icons = [];
		let settings = this.settings = Settings.cache[id];
		this.id = id;
		//this.settings.fav_label = false;
		
		this.actor = new St.BoxLayout({ name: "panelDocklet",
				style_class: "panelDocklet",
				reactive: true
			});
		this.windowActor = new St.BoxLayout({
				style_class: "windowActor"
			});
		this.windowActor_vertical = new St.BoxLayout({x_expand:true, y_expand:true});
		this.actor.add(this.windowActor_vertical);
		this.windowActor_vertical.add(this.windowActor);
		
		
		//correct settings
		if(settings.dock_monitor >= LayoutManager.monitors.length)
			settings.dock_monitor = LayoutManager.monitors.length-1;
		
		if(settings.icons_overhang > settings.icon_size - MIN_OVERHANG_DIFF)
			settings.icons_overhang = settings.icon_size - MIN_OVERHANG_DIFF;
		
		if(settings.filter_workspaces)
			settings.icon_ws_indicator = false;
		if(settings.filter_monitors)
			settings.icon_monitor_indicator = false;
		
		//positioning and sizing
		//
		let monitor = LayoutManager.monitors[settings.dock_monitor];
		
		
		let item;
		switch(settings.dock_orientation) {
			case DOCK_POS.PANEL_LEFT:
				Main.panel._leftBox.insert_child_at_index(this.actor, 1);
				this.intoPanel = true;
				this.icon_space = monitor.width/2 - 75; //will be corrected in solve_shaking()
			break;
			case DOCK_POS.PANEL_MIDDLE_LEFT:
				Main.panel._centerBox.insert_child_at_index(this.actor, 0);
				this.intoPanel = true;
				this.icon_space = monitor.width - 225; //will be corrected in solve_shaking()
			break;
			case DOCK_POS.PANEL_MIDDLE_RIGHT:
				Main.panel._centerBox.insert_child_at_index(this.actor, 1);
				this.intoPanel = true;
				this.icon_space = monitor.width - 225; //will be corrected in solve_shaking()
			break;
			case DOCK_POS.PANEL_RIGHT:
				Main.panel._rightBox.insert_child_at_index(this.actor, 0);
				this.intoPanel = true;
				this.icon_space = monitor.width/2; //will be corrected in solve_shaking()
			break;
			case DOCK_POS.TOP:
				this.actor.set_y(this._dock_y = monitor.y);
				this.actor.add_style_class_name("top");
				this.menu = new RightClickMenu(this, St.Side.TOP);
				
				this._dock_hide_y = monitor.y - settings.icon_size + settings.icons_overhang + 1; //WARNING: it needs to be 1px inside!
			break;
			case DOCK_POS.RIGHT:
				this.actor.set_x(this._dock_x = monitor.x + monitor.width - settings.icon_size + settings.icons_overhang - DOCK_PADDING_SHORT_SUM);
				this.actor.add_style_class_name("right");
				this.menu = new RightClickMenu(this, St.Side.RIGHT);
				this.windowActor_vertical.set_pack_start(true);
				
				this._dock_hide_x = monitor.x + monitor.width - 1; //WARNING: it needs to be 1px inside!
				this.vertical = true;
			break;
			case DOCK_POS.BOTTOM:
				this.actor.set_y(this._dock_y = monitor.y + monitor.height - settings.icon_size + settings.icons_overhang - DOCK_PADDING_SHORT_SUM);
				this.actor.add_style_class_name("bottom");
				this.menu = new RightClickMenu(this, St.Side.BOTTOM);
				this.windowActor_vertical.set_pack_start(true);
				
				this._dock_hide_y = monitor.y + monitor.height - 1; //WARNING: it needs to be 1px inside!
			break;
			case DOCK_POS.LEFT:
				this.actor.set_x(this._dock_x = monitor.x);
				this.actor.add_style_class_name("left");
				this.menu = new RightClickMenu(this, St.Side.LEFT);
				
				this._dock_hide_x = monitor.x - settings.icon_size + settings.icons_overhang + 1; //WARNING: it needs to be 1px inside!
				this.vertical = true;
			break;
		}
		
		
		//complete sizing (we needed to wait for extras)
		//
		if(this.intoPanel) {
			if(settings.dock_fixed_size) {
				this.icon_space = settings.dock_size - DOCK_PADDING_LONG_SUM;
				this.actor.set_width(settings.dock_size);
			}
				
			
			this.windowActor_vertical.set_vertical(true);
			this.actor.add_style_class_name("panel top");
			this.menu = new RightClickMenu(this, St.Side.TOP);
			
			settings.icons_overhang = Math.max(settings.icon_size - Main.panel.actor.get_height(), 0);
			settings.dock_align = St.Align.START; //must not be anything else when in panel 
			this._dock_hide_y = this._dock_hide_x = this._dock_y = this._dock_x = 0;
		}
		else {
			LayoutManager.addChrome(this.actor, {trackFullscreen:true, affectsStruts: settings.affect_strut});
			if(this.vertical) {
				this.actor.set_y(this._dock_hide_y = this._dock_y = monitor.y + settings.dock_position);
				if(settings.dock_fixed_size) {
					this.icon_space = settings.dock_size - DOCK_PADDING_LONG_SUM;
					this.actor.set_size(settings.icon_size + DOCK_PADDING_SHORT_SUM - settings.icons_overhang, settings.dock_size);
				}
				else {
					this.actor.set_width(settings.icon_size + DOCK_PADDING_SHORT_SUM - settings.icons_overhang);
					switch(settings.dock_align) {
						case St.Align.END:
							this.icon_space = settings.dock_position - DOCK_PADDING_LONG_SUM;
						break;
						case St.Align.MIDDLE:
							this.icon_space = Math.min(settings.dock_position, monitor.height - settings.dock_position)*2 - DOCK_PADDING_LONG_SUM;
						break;
						default:
							this.icon_space = monitor.height - settings.dock_position - DOCK_PADDING_LONG_SUM;
					}
				}
				this.actor.set_vertical(true);
				this.windowActor.set_vertical(true);
			}
			else {
				this.actor.set_x(this._dock_hide_x = this._dock_x = monitor.x + settings.dock_position);
				if(settings.dock_fixed_size) {
					this.icon_space = settings.dock_size - DOCK_PADDING_LONG_SUM;
					this.actor.set_size(settings.dock_size, settings.icon_size + DOCK_PADDING_SHORT_SUM - settings.icons_overhang);
				}
				else {
					this.actor.set_height(settings.icon_size + DOCK_PADDING_SHORT_SUM - settings.icons_overhang);
					
					switch(settings.dock_align) {
						case St.Align.END:
							this.icon_space = settings.dock_position - DOCK_PADDING_LONG_SUM;
						break;
						case St.Align.MIDDLE:
							this.icon_space = Math.min(settings.dock_position, monitor.width - settings.dock_position)*2 - DOCK_PADDING_LONG_SUM;
						break;
						default:
							this.icon_space = monitor.width - settings.dock_position - DOCK_PADDING_LONG_SUM;
					}
				}
				this.actor.set_x(this._dock_hide_x = this._dock_x = monitor.x + settings.dock_position);
				this.windowActor_vertical.set_vertical(true);
			}
		}
		
		if(settings.affect_strut)
			this.actor.add_style_class_name("strut");
		
		this.label_length = settings.label_length; //will be changed
		this.icon_size = settings.icon_size; //will be changed
		
		
		//Rightclickmenu
		//
		this.menu_manager = new PopupMenu.PopupMenuManager(this);
        this.menu_manager.addMenu(this.menu);
        Main.uiGroup.add_actor(this.menu.actor);
        this.menu.actor.hide();
        
		
		//listeners
		//
		this._ID_overviewShown = Main.overview.connect("showing", Lang.bind(this, this.unvanish));
		this._ID_overviewHidden = Main.overview.connect("hiding", Lang.bind(this, this.vanish));
		
		if(settings.autohide) {
			this.actor.connect("enter-event", Lang.bind(this, this.show));
			this.actor.connect("leave-event", Lang.bind(this, this.hide));
			this._ID_xdnd_drag_begin = Main.xdndHandler.connect("drag-begin", Lang.bind(this, this.show));
			this._ID_xdnd_drag_end = Main.xdndHandler.connect("drag-end", Lang.bind(this, this.hide));
			this.hide();
		}
		
		//workspace_indicator
		//
		if(settings.workspace_indicator) {
			this.workspaceActor = new St.BoxLayout({
				style_class: "workspaceActor",
				vertical: this.vertical
			});
			
			if(settings.workspace_indicator_align == St.Align.START)
				this.actor.insert_child_at_index(this.workspaceActor, 0);
			else
				this.actor.add(this.workspaceActor);
			
			this.icon_space -= WORKSPACE_INDICATOR.PADDING_SUM;
			this.extra_space += WORKSPACE_INDICATOR.PADDING_SUM;
		}
		
		//tray_button
		//
		if(settings.tray_button) {
			let tray_button = new St.Button({
					style_class: "panel-button",
					y_align: St.Align.END
				});
			tray_button.set_child(new St.Icon({
					icon_name: "preferences-system-notifications-symbolic",
					style_class: "tray_button",
					width: 26,
					height: 16
				}));
			tray_button.connect("button-release-event", function() {
					if(Main.messageTray._trayState == MessageTray.State.HIDDEN)
						Main.messageTray.openTray();
					else
						Main.messageTray.close();
				});
			
			if(settings.tray_button_align == St.Align.START)
				this.actor.insert_child_at_index(tray_button, 0);
			else
				this.actor.add(tray_button);
				//this.actor.add(tray_button, {y_fill:false, x_fill:false});
			
			this.icon_space -= settings.icon_size;
			this.extra_space += settings.icon_size;
		}
		
		//Hover-menu
		//
		if(settings.hover_menu) {
			let box = new St.BoxLayout(obj_size({
					vertical: this.vertical,
					translation_x: this.vertical ? (settings.dock_orientation == DOCK_POS.LEFT ? -settings.icons_overhang : settings.icons_overhang) : 0,
					translation_y: (!this.vertical) ? (settings.dock_orientation == DOCK_POS.BOTTOM ? settings.icons_overhang : -settings.icons_overhang) : 0,
				}, this.vertical, 0)); //so windowActor wont expand with hover_box width
			
			this.hover_box = new St.BoxLayout({
				vertical: true,
				reactive: true,
				track_hover: true,
				style_class: "hover_menu"
			});
			this.hover_box.hide();
			this.hover_box.opacity = 0;
			
			box.add(this.hover_box);
			this.windowActor_vertical.add(box, {x_align: St.Align.START, y_align: St.Align.START, x_fill: false, y_fill: false});
			LayoutManager.trackChrome(this.hover_box);
			this.actor.connect("leave-event", Lang.bind(this, this.close_hover_menu));
		}
		
        //import data
        //
		this.update_icon_size(); //has to be done after extras
		Database.request_icons(this);
		this._import_data();
	},
	
	open_hover_menu: function() {
		Tweener.removeTweens(this.hover_box);
		if(this._hover_timeout) {
			Mainloop.source_remove(this._hover_timeout);
			this._hover_timeout = false;
		}
		
		this.hover_box.show();
		Tweener.addTween(this.hover_box, {
				opacity: 255,
				time: 0.4,
				transition: "easeOutQuint"
			});
		this.hover_menu_isOpen = true;
	},
	close_hover_menu: function(actor) {
		if(this._hover_timeout)
			Mainloop.source_remove(this._hover_timeout);
		
		this._hover_timeout = Mainloop.timeout_add(100, Lang.bind(this, function() {
				if(this.hover_box.get_hover()) {
					if(!this.hover_menu_isOpen)
						this.open_hover_menu();
				}
				else
					this._do_close_hover_menu();
			}));
	},
	_do_close_hover_menu: function(actor) {
		this._hover_timeout = false;
		
		Tweener.removeTweens(this.hover_box);
		Tweener.addTween(this.hover_box, {
				opacity: 0,
				time: 0.4,
				transition: "easeOutQuint",
				onComplete: function() {this.hide();}
			});
		this.hover_menu_isOpen = false;
	},
	
	index_favs: function(i) {
		let favs = AppFavorites.getAppFavorites().getFavorites(),
			monitor = this.settings.filter_monitors_num,
			favBox = (this.fav_icons[i] = {});
		favBox.l = favs.length;
		
		for(let k in favs) {
			favBox[favs[k].get_id()] = this.create_fav_icon(favs[k], i);
		}
	},
	create_fav_icon: function(app, ws_index) {
		let obj = Object.create(FAV_WIN_MASC);
		obj.window.title = app.get_name();
		
		let icon = new Icon(obj, this, app, ws_index, this.settings.fav_label);
		this.insert_icon(icon, ws_index);
		icon.mark_fav_icon();
		
		return icon;
	},
	update_favs: function() {
		if(!this.settings.favorites)
			return;
		
		let fav_obj = AppFavorites.getAppFavorites(),
			favs = fav_obj.getFavorites(),
			favBox = this.fav_icons[0];
			
		if(favs.length > favBox.l) {//new fav
			let new_fav;
			
			for(let k in favs) {//find fav which isnt indexed yet
				new_fav = favs[k];
				if(!favBox.hasOwnProperty(new_fav.get_id())) {
					break;
				}
			}
			
			
			let win_len = new_fav.get_windows().length,
				i = Database.workspaces.length,
				name = new_fav.get_id();
			
			
			if(!win_len) { //no windows exist
				while(i--) {
					++this.fav_icons[i].l;
					this.fav_icons[i][name] = this.create_fav_icon(new_fav, i, i == this.currentWS);
				}
			}
			else { //at least one window exists
				while(i--) {
					++this.fav_icons[i].l;
					this.fav_icons[i][name] = false;
				}
			}
			
		}
		else {//deleted fav
			let name;
			for(name in favBox) { //find deleted fav
				if(name != "l" && !fav_obj.isFavorite(name))
					break;
			}
			let i = Database.workspaces.length;
			while(i--) {
				if(this.fav_icons[i][name]) //icon is still in fav-mode - we can remove it
					this.fav_icons[i][name].destroy(true, true);
				--this.fav_icons[i].l;
				delete this.fav_icons[i][name];
			}
		}
	},
	
	_import_data: function() {
		let ws = Database.workspaces,
			wins = Database.windows,
			winBox, metaWin;
		
		//create workspaces first
		if(this.settings.filter_workspaces) {
			for(let ws_index in ws) {
				this.add_ws(ws_index, []);
			}
		}
		else {
			if(this.settings.favorites)
				this.index_favs(0);
			this.applications[0] = {};
		}
		if(this.settings.workspace_indicator)
			this.select_ws_button(0, global.screen.get_active_workspace_index());
		
		for(let k in wins) {
			winBox = wins[k];
			metaWin = winBox.window;
			this.create_icon_check(winBox, metaWin.get_workspace().index(), metaWin.get_monitor());
			if(metaWin.is_on_all_workspaces())
				this.set_all_ws_win(winBox, metaWin.get_workspace().index(), metaWin.get_monitor());
		}
		
		this.currentWS = this.settings.filter_workspaces ? global.screen.get_active_workspace_index() : 0;
		this.to_ws(this.currentWS);
	},
	
	remove_ws: function(removedIndex, removedNum) {
		if(!this.settings.filter_workspaces)
			return;
		
		let i = removedIndex + removedNum,
			max=this.windowList.length,
			t, k;
		
		while(i-- > removedIndex) {
			t = this.windowList[i];
			while(t) {
				t.destroy(true);
				t = t.next;
			}
			
			//settings.only_apps:
			t = this.applications[i];
			for(k in t) {
				t[k].destroy(true);
			}
			this.applications.splice(removedIndex, removedNum);
		}
		if(this.settings.favorites)
			this.fav_icons.splice(removedIndex, removedNum);
		if(this.settings.workspace_indicator)
			this.remove_ws_button(removedIndex, removedNum);
		
		
		//correct ws_index of icons
		for(i=removedIndex+removedNum; i<max; ++i) {
			t = this.windowList[i];
			while(t) {
				t.ws_index -= removedNum;
				t = t.next;
			}
		}
		
		this.currentWS = global.screen.get_active_workspace_index(); //has to happen after icons are removed
		this.windowList.splice(removedIndex, removedNum);
	},
	add_ws: function(id, all_ws_winBoxes) {
		if(this.settings.workspace_indicator)
			this.create_ws_button(id);
		
		if(!this.settings.filter_workspaces)
			return;
		
		this.windowList[id] = false;
		this.applications[id] = {};
		
		if(this.settings.favorites)//add favs in new ws and create array- has to happen before icons are added
			this.index_favs(id);
		
		for(let k in all_ws_winBoxes) {
			this.create_icon_check(all_ws_winBoxes[k], id, all_ws_winBoxes[k].window.get_monitor());
		}
		
	},
	switch_ws: function(from, to) {
		if(this.settings.workspace_indicator)
			this.select_ws_button(from, to);
		
		if(!this.settings.filter_workspaces) {
			let icon = this.windowList[0];
			
			while(icon) {
				icon.mark_ws();
				//icon.mark_ws(to);
				icon = icon.next;
			}
			
			
			return;
		}
		
		let list = this.windowList[from],
			winBox;
		
		while(list) {
			winBox = list.winBox;
			
			if(winBox.window.is_on_all_workspaces()) {
				if(!winBox.all_ws)
					winBox.set_all_ws(from);
			}
			else if(winBox.all_ws) {// && !winBox.window.is_on_all_workspaces()
				winBox.unset_all_ws();
			}
			list.remove();
			list = list.next;
		}
		this.to_ws(to);
	},
	to_ws: function(to) {
		this.currentWS = to;//has to be set after old ws has processed or all_ws_windows will be inserted twice
		this.iconList_count = this.iconLabel_count = 0; //we will recalc now!
		let list = this.windowList[to]; //there could have been changes to to-list, so we need to load it after from-list is processed
		while(list) {
			if(list._label)
				++this.iconLabel_count;
			++this.iconList_count;
			list.insert_no_size();
			list = list.next;
		}
		this.update_icon_size();
	},
	win_has_changed_ws: function(winBox, icon, to) {
		if(!this.settings.filter_workspaces)
			return;
		
		icon.destroy(false, false, winBox);
		this.create_icon(winBox, to, winBox.window.get_monitor(), this.settings.only_apps);
	},
	
	remove_ws_button: function(i, num) {
		while(num--) {
			this.workspaceActor.get_child_at_index(i+num).destroy();
		}
		let diff = this.settings.icon_size + WORKSPACE_INDICATOR.BUTTON_MARGIN + WORKSPACE_INDICATOR.BUTTON_PADDING_SUM;
		this.icon_space += diff;
		this.extra_space -= diff;
		this.update_icon_size();
	},
	create_ws_button: function(id) {
		let margin = WORKSPACE_INDICATOR.BUTTON_MARGIN,
			obj = {
					margin_left: margin,
					margin_right: margin,
					margin_top: margin,
					margin_bottom: margin,
				};
		
		if(this.vertical) {
			if(this.settings.dock_orientation == DOCK_POS.LEFT)
				obj.margin_left = 0;
			else
				obj.margin_right = 0;
			obj.height = this.settings.icon_size;
		}
		else {
			if(this.settings.dock_orientation == DOCK_POS.TOP)
				obj.margin_top = 0;
			else
				obj.margin_bottom = 0;
			obj.width = this.settings.icon_size;
		}
		let button = new St.Button(obj),
			ws = global.screen.get_workspace_by_index(id); //for connect-function
		
		button.connect("button-release-event", function() {
				ws.activate(global.get_current_time());
			});
		//button.set_child(new St.Label({text: id.toString()}));
		
		this.workspaceActor.add(button);
		
		let diff = this.settings.icon_size + margin*2 + WORKSPACE_INDICATOR.BUTTON_PADDING_SUM;
		this.icon_space -= diff;
		this.extra_space += diff;
		
		this.update_icon_size();
	},
	select_ws_button: function(from, to) {
		let child_from = this.workspaceActor.get_child_at_index(from),
			child_to = this.workspaceActor.get_child_at_index(to);
		
		child_from.remove_style_class_name("current");
		child_to.add_style_class_name("current");
	},
	
	remove_icon: function(icon) {
		if(icon.ws_index == this.currentWS) {
			--this.iconList_count;
			if(icon._label)
				--this.iconLabel_count;
			Mainloop.timeout_add(200, Lang.bind(this, function() {this.update_icon_size();})); //to make sure actors arent expanded and shrinked at the same time -> shaking
		}
		
		if(!icon.previous) {//meaning dock.windowList[x] == this -> lets find x
			//workspace.index()-information is not relyable because this icon could be a all_ws-copy:
			let list = this.windowList;
			
			for(let k in list) {
				if(list[k] == icon) {
					list[k] = icon.next;
					icon.next.previous = false;
					break;
				}
			}
		}
		else {
			icon.previous.next = icon.next;
			if(icon.next)
				icon.next.previous = icon.previous;
		}
	},
	create_icon: function(winBox, ws_index, monitor, only_apps) {
		let real_ws = ws_index;
		if(!this.settings.filter_workspaces)
			ws_index = 0;
		
		if(this.settings.favorites) {
			/* settings.favorites:
				one application-entry in fav_icons per ws
				if entry is false, than there is at least one window in docklet
				otherwise its a pointer to its fav-icon
				fav_icons[x].l counts the number of indexed application-favs
				
				the first window of a fav-application becomes an fav-icon
				when the last application-window is removed, its icon will be turned into fav-mode
			*/
			let favs = this.fav_icons[ws_index],
				app_name = winBox._app.get_id(),
				fav;
			if((fav = favs[app_name]) && fav.is_fav_icon) {
				favs[app_name] = false;
				fav.unmark_fav_icon(winBox);
				
				if(only_apps)
					fav.set_app_icon(winBox._app.get_id());
				fav.mark_monitor(monitor);
				fav.mark_ws(real_ws);
				
				return fav;
			}
		}
		
		let app_name = winBox._app.get_id(),
			icon;
		
		if(this.applications[ws_index].hasOwnProperty(app_name)) {
			/* settings.only_apps:
				save the first window-icon and increase its app_windows for every following window from the same app.
				icon will be removed only when app_windows==0
			*/
			icon = this.applications[ws_index][app_name];
			winBox.index_icon(icon); //when window gets closed, collectors-icon need to be informed about it - and decrease app_windows
			icon.increase_app_windows(winBox);
			return icon;
		}
		else
			icon = winBox.new_icon(this, ws_index);
		
		if(only_apps)
			icon.set_app_icon(app_name);
		icon.mark_monitor(monitor);
		icon.mark_ws(real_ws);
		
		this.insert_icon(icon, ws_index);
		return icon;
	},
	insert_icon: function(icon, ws_index) {
		let insert = ws_index == this.currentWS,
			last;
		if(this.settings.ordered) {
			last = this.get_first_app_window(this.windowList[ws_index], icon.app);
			
			//we want to get the last one of the pack:
			while(last.next && last.next.app == icon.app) {
				last = last.next;
			}
			
			if(last.next) {
				icon.next = last.next;
				last.next.previous = icon;
			}
		}
		else
			last = this.get_last_window(ws_index);
		
		if(!last) {
			this.windowList[ws_index] = icon;
		}
		else {
			last.next = icon;
			icon.previous = last;
		}
		
		if(insert) {
			++this.iconList_count;
			if(icon._label)
				++this.iconLabel_count;
			if(this.update_icon_size())
				icon.insert_timeout(last.actor); //make sure actors arent expanded and shrinked at the same time -> shaking
			else
				icon.insert(last.actor);
		}
	},
	create_icon_check: function(winBox, ws_index, monitor) {
		if(this.settings.filter_monitors && monitor != this.settings.filter_monitors_num)
			return;
		
		this.create_icon(winBox, ws_index, monitor, this.settings.only_apps);
	},
	update_icon_size: function() {
		let icon_size_sum = this.iconList_count * this.icon_size;
		if(this.iconLabel_count * this.label_length + icon_size_sum > this.icon_space)
			this.iconList_is_decreased = true;
		else if(!this.iconList_is_decreased) {//false: means, that before there wasnt enough space for icons and now there is
			this.update_dock_position();
			return false;
		}
		let icon = this.windowList[this.currentWS],
			new_label_length = this.iconLabel_count ? //this.iconLabel_count can be 0!
					Math.min(this.settings.label_length, Math.round((this.icon_space - icon_size_sum) / this.iconLabel_count)) : 0;
		
		if(new_label_length >= MIN_LABEL_LENGTH) {
			this.label_length = new_label_length;
			
			while(icon) {
				if(icon._label)
					icon.change_label_size(new_label_length);
				icon = icon.next;
			}
			if(new_label_length == this.settings.label_length)
				this.iconList_is_decreased = false;
		}
		else {
			let new_icon_size = this.iconList_count ? Math.round(this.icon_space / this.iconList_count) : this.icon_space; //this.iconList_count can be 0!
			//make sure no actor is smaller than 1 (especialy Icon._icon because of scale in Icon.mark_fav_icon() -> crash!)
			this.icon_size = ((new_icon_size <= MIN_ICON_SIZE) ? MIN_ICON_SIZE : Math.min(new_icon_size, this.settings.icon_size));
			
			while(icon) {
				icon.change_icon_size(this.icon_size);
				icon = icon.next;
			}
			if(this.icon_size == this.settings.icon_size && !this.settings.icon_label)
				this.iconList_is_decreased = false;
		}
		this.update_dock_position();
		return true;
	},
	update_dock_position: function() {
		if(this.settings.dock_fixed_size)
			return;
		
		let value;
		switch(this.settings.dock_align) {
			case St.Align.MIDDLE:
				value = 2;
			break;
			case St.Align.END:
				value = 1;
			break;
			default:
				return;
		}
		//let size = this.actor.get_transformed_size(); //WARNING: asuming, that this gets calculated before other Tweeners have started
		
		let size = this.extra_space + this.iconList_count * this.icon_size + this.iconLabel_count * this.label_length;
		Tweener.addTween(this.actor, obj_translation({
				time: 0.4,
				transition: "easeOutSine"
			}, this.vertical, -size/value));
	},
	win_left_monitor: function(metaWin, monitor, ws) {
		if(!this.settings.filter_monitors || this.settings.filter_monitors_num != monitor)
			return;
		//if metaWin.is_on_all_workspaces() icon could alco be a is_all_ws_clone - but we dont care because we will destroy all anyway
		let ws_index = this.settings.filter_workspaces ? metaWin.get_workspace().index() : 0,
			app_name = get_app(metaWin).get_id(),
			icon;
		
		if(this.applications[ws_index].hasOwnProperty(app_name)) //settings.only_apps or manualy set to app-icon
			icon = this.applications[ws_index][app_name];
		else {
			icon = this.windowList[ws_index]
			while(icon) {
				if(icon.winBox.window == metaWin)
					break;
				icon = icon.next;
			}
		}
		
		if(!icon) 
			return;
		if(icon.winBox.all_ws) {
			let icons = icon.winBox.icons,
				i = icons.length,
				winBox = icon.winBox;
			
			while(i--) {//WARNING: there will be splice, so removal has to start from back!
				if(icons[i].dock == this)
					icons[i].destroy(false, false, winBox);
			}
		}
		else
			icon.destroy(false, false, icon.winBox);
	},
	win_entered_monitor: function(winBox, monitor) {
		if(!this.settings.filter_monitors || this.settings.filter_monitors_num != monitor)
			return;
		
		let ws_index = winBox.window.get_workspace().index();
		this.create_icon(winBox, ws_index, monitor, this.settings.only_apps);
		
		if(winBox.all_ws) {
			this.set_all_ws_win(winBox, ws_index, monitor);
		}
	},
	
	set_all_ws_win: function(winBox, ws_index_to_skip, monitor) {
		if(!this.settings.filter_workspaces || (this.settings.filter_monitors && monitor != this.settings.filter_monitors_num))
			return;
		
		let windowList = this.windowList,
			only_apps = this.settings.only_apps;
		for(let k in windowList) {
			if(k == ws_index_to_skip)
				continue;
			
			this.create_icon(winBox, k, monitor, only_apps).is_all_ws_clone = true;
		}
	},
	
	get_last_window: function(i) {
		let win = this.windowList[i];
		
		if(!win)
			return false;
		
		while(win.next) {
			win = win.next;
		}
		return win;
	},
	get_first_app_window: function(win, app) {
		if(!win)
			return false;
		
		while(win.next) {
			if(win.app == app)
				break;
			win = win.next;
		}
		return win;
	},
	
	select: function() {
		if(this.hidden)
			this.show();
		
		this.selected = true;
		this.actor.add_style_pseudo_class("selected");
	},
	deselect: function() {
		if(this.settings.autohide)
			this._do_hide();
		this.selected = false;
		this.actor.remove_style_pseudo_class("selected");
	},
	show: function() {
		if(this._hide_timeout) {
			Mainloop.source_remove(this._hide_timeout);
			this._hide_timeout = false;
		}
		this.hidden = false;
		Tweener.addTween(this.actor, {
				x: this._dock_x,
				y: this._dock_y,
				opacity: 255,
				time: 0.2,
				transition: "easeOutQuint"
			});
	},
	hide: function() {
		if(this.selected || this.menu.isOpen)
			return;
		this.hidden = true;
		if(this._hide_timeout) {
			Mainloop.source_remove(this._hide_timeout);
			this._hide_timeout = false;
		}
		this._hide_timeout = Mainloop.timeout_add(this.settings.hide_timeout_ms, Lang.bind(this, this._do_hide));
	},
	_do_hide: function() {
		this.hidden = true;
		this._hide_timeout = false;
		Tweener.addTween(this.actor, {
				x: this._dock_hide_x,
				y: this._dock_hide_y,
				opacity: this.settings.hide_opacity,
				time: 0.4,
				transition: "easeOutQuint"
			});
	},
	vanish: function() {
		this.actor.show();
	},
	unvanish: function() {
		this.actor.hide();
	},
	
	destroy: function(no_id_correct) {
		Main.overview.disconnect(this._ID_overviewShown);
		Main.overview.disconnect(this._ID_overviewHidden);
		if(this._ID_xdnd_drag_begin) {
			Main.xdndHandler.disconnect(this._ID_xdnd_drag_begin);
			Main.xdndHandler.disconnect(this._ID_xdnd_drag_end);
		}
		
		Database.remove_dock(this, no_id_correct);//this docklets icons will also get a destroy()
		
		this.actor.destroy();
	}
});

const Database = {
    workspaces: null, //array - is only needed to check which ws has been removed
    windows: null, //object
    
    connected_docks: null, //array
    all_ws_wins: null, //array
    last_focus: false, //will be used from Window
    
    _ID_newWorkspace: null,
    _ID_switchWorkspace: null,
    _ID_enteredMonitor: null,
    _ID_leftMonitor: null,
    _ID_monitorEvent: null,
    _ID_windowAdded: null,
    _ID_minimize: null,
    _ID_map: null,
    _ID_update_favs: null,
    
    init: function() {
		this.workspaces = [];
		this.windows = {};
		this.connected_docks = [];
		this.all_ws_wins = [];
		
		this._ID_newWorkspace = global.screen.connect("notify::n-workspaces", Lang.bind(this, this._on_ws_modification));
		this._ID_switchWorkspace = global.window_manager.connect("switch-workspace", Lang.bind(this, this._on_switch_ws));
		this._ID_enteredMonitor = global.screen.connect("window-entered-monitor", Lang.bind(this, this._on_entered_monitor));
		this._ID_leftMonitor = global.screen.connect("window-left-monitor", Lang.bind(this, this._on_left_monitor));
		this._ID_monitorEvent = global.screen.connect("monitors-changed", Lang.bind(this, this._on_monitors_changed));
		this._ID_windowAdded = global.display.connect("window-created", Lang.bind(this, this._on_window_added));
		this._ID_attention = global.display.connect("window-demands-attention", Lang.bind(this, this._on_window_attention));
		this._ID_minimize = global.window_manager.connect("minimize", Lang.bind(this, this._on_minimize));
		this._ID_map = global.window_manager.connect("map", Lang.bind(this, this._on_map));
		
		this._ID_update_favs = AppFavorites.getAppFavorites().connect("changed", Lang.bind(this, this._on_update_favs));
		
		//TODO: monitors changed
		
		this.fill_db();
	},
	fill_db: function() {
		let ws_num = global.screen.n_workspaces,
			wins;
		try {		wins = global.display.get_tab_list(Meta.TabList.NORMAL, null);} //gnome-shell 3.16
		catch(e) {	wins = global.display.get_tab_list(Meta.TabList.NORMAL, global.screen, null);}
		
		while(ws_num--) {
			this.workspaces[ws_num] = global.screen.get_workspace_by_index(ws_num);
		}
		
		for(let k in wins) {
			this._on_window_added(null, wins[k]); //WARNING: asuming, that we wont need metaDisplay
		}
	},
	
	insert_win: function(metaWin) {
		let docks = this.connected_docks,
			monitor = metaWin.get_monitor(),
			winBox = new Window(metaWin),
			ws_index = metaWin.get_workspace().index();
		
		this.windows[metaWin.get_stable_sequence()] = winBox;
		
		for(let k in docks) {
			docks[k].create_icon_check(winBox, ws_index, monitor);
		}
		return winBox;
	},
	remove_win: function(metaWin) {
		let id = metaWin.get_stable_sequence();
		this.windows[id].destroy();
		delete this.windows[id];
	},
	index_all_ws_win: function(winBox) {
		this.all_ws_wins.push(winBox);
	},
	unindex_all_ws_win: function(winBox) {
		let all_ws_wins = this.all_ws_wins;
		
		for(let k in all_ws_wins) {
			if(all_ws_wins[k] == winBox) {
				all_ws_wins.splice(k, 1);
				break;
			}
		}
	},
	
	request_icons: function(dock) {
		this.connected_docks[dock.id] = dock;
	},
	remove_dock: function(dock, no_id_correct) {
		let icons, k, i, max;
		
		this.connected_docks.splice(dock.id, 1);
		//correct ids:
		if(!no_id_correct) {
			for(i=k, max = this.connected_docks.length; i<max; ++i) {
				this.connected_docks[i].id = i;
			}
		}
		
		
		for(k in this.windows) {
			icons = this.windows[k].icons;
			i = icons.length;
			while(i--) {//we have to start in the back because array will be shrinked!
				if(icons[i].dock == dock) 
					this.windows[k].remove_icon(i);
			}
		}
	},
	
	check_ws: function(ws_index) {
		//NOTE: creating a window on an empty workspace sets it to on_all_ws first - we have to ignore it until window-added-event is triggered. So no checking for windows
		let docks = this.connected_docks,
			all_ws_wins = Database.all_ws_wins;
		
		for(let k in docks) {
			docks[k].add_ws(ws_index, all_ws_wins);
		}
	},
	_on_ws_modification: function() {
		//code-parts from: usr/share/gnome-shell/js/ui/main.js
		let oldNumWorkspaces = this.workspaces.length,
			newNumWorkspaces = global.screen.n_workspaces;
		
		if(newNumWorkspaces > oldNumWorkspaces) {
			// Assume workspaces are only added at the end
			for(let i = oldNumWorkspaces; i < newNumWorkspaces; ++i) {
				this.workspaces[i] = global.screen.get_workspace_by_index(i);
				this.check_ws(i);
			}
		}
		else if(newNumWorkspaces < oldNumWorkspaces) {
			// Assume workspaces are only removed sequentially
			// (e.g. 2,3,4 - not 2,4,7)
			let removedNum = oldNumWorkspaces - newNumWorkspaces,
				removedIndex;
			for(let i=0; i < oldNumWorkspaces; ++i) {
				if(this.workspaces[i] != global.screen.get_workspace_by_index(i)) {
					removedIndex = i;
					break;
				}
			}
			for(let k in this.connected_docks) {
				this.connected_docks[k].remove_ws(removedIndex, removedNum);
			}
			this.workspaces.splice(removedIndex, removedNum);
		}
	},
	_on_switch_ws: function(wm, from, to, direction) {
		let docks = this.connected_docks;
		
		for(let k in docks) {
			docks[k].switch_ws(from, to);
		}
	},
	_on_entered_monitor: function(metaScreen, monitor, metaWin) {
		let winBox = this.windows[metaWin.get_stable_sequence()];
		if(!winBox)
			return;
		
		winBox.moved_to_monitor(monitor);
		
		for(let k in this.connected_docks) {
			this.connected_docks[k].win_entered_monitor(winBox, monitor);
		}
	},
	_on_left_monitor: function(metaScreen, monitor, metaWin) {
		let ws = metaWin.get_workspace();
		
		if(!metaWin.get_compositor_private()) //for some reason, left monitor kinda randomly fires when window is closed (always) or shell is reloaded (sometimes)
			return
		
		if(this.windows[metaWin.get_stable_sequence()]) //lets filter windows we dont know/ care about
			for(let k in this.connected_docks) {
				this.connected_docks[k].win_left_monitor(metaWin, monitor, ws);
			}
	},
	_on_monitors_changed: function() {
		let i = Settings.get_int("dock-count");
		while(i--) {
			this.connected_docks[i].destroy();
			this.connected_docks[i] = new Docklet(i);
		}
	},
	_on_minimize: function(shellwm, actor) {
		let metaWin = actor.get_meta_window(),
			winBox = this.windows[metaWin.get_stable_sequence()];
		
		if(winBox)
			winBox.set_minimized();
	},
	_on_map: function(shellwm, actor) {
		let metaWin = actor.get_meta_window(),
			winBox = this.windows[metaWin.get_stable_sequence()];
		
		if(winBox)
			winBox.unset_minimized();
	},
	_on_update_favs: function(appFavs) {
		for(let k in this.connected_docks) {
			this.connected_docks[k].update_favs(appFavs);
		}
	},
	_on_window_attention: function(display, metaWin) {
		if(!metaWin || metaWin.has_focus() || metaWin.is_skip_taskbar()) //from windowAttentionHandler.js:32
			return;
		
		let winBox = this.windows[metaWin.get_stable_sequence()];
		if(!winBox)
			return;
		winBox.set_urgent();
	},
	_on_window_added: function(metaDisplay, metaWin) { //WARNING: other functions asume, that we wont need metaDisplay
		if(metaWin.is_skip_taskbar())
			return;
		
		let winBox = this.insert_win(metaWin);
		if(metaWin.is_on_all_workspaces())
			winBox.set_all_ws(metaWin.get_workspace().index());
	},
	
	destroy: function() {
		global.screen.disconnect(this._ID_newWorkspace);
		global.window_manager.disconnect(this._ID_switchWorkspace);
		global.screen.disconnect(this._ID_enteredMonitor);
		global.screen.disconnect(this._ID_leftMonitor);
		global.screen.disconnect(this._ID_monitorEvent);
		global.display.disconnect(this._ID_windowAdded);
		global.display.disconnect(this._ID_attention);
		global.window_manager.disconnect(this._ID_minimize);
		global.window_manager.disconnect(this._ID_map);
		AppFavorites.getAppFavorites().disconnect(this._ID_update_favs);
		
		for(let i in this.windows) {
			this.windows[i].destroy();
		}
		
		//This is a global object! Lets shrink it down:
		this.workspaces = null;
		this.windows = null;
		this.connected_docks = null;
		this.all_ws_wins = null;
	}
};
const Window = new Lang.Class({
	Name: "Window",

	window: null,
	_app: null,
	all_ws: false,
	closed: false,
	icons: null, //array
	
	_ID_unmanaged: null,
	_ID_focus: null,
	_ID_title: null,
	_ID_ws_changed: null,
	
	_init: function(metaWin) {
		this.icons = [];
		this.window = metaWin;
		this._app = get_app(metaWin);
		
		this._ID_unmanaged = metaWin.connect("unmanaged", Lang.bind(Database, Database.remove_win));
		this._ID_focus = metaWin.connect("focus", Lang.bind(this, this.set_focus));
		this._ID_title = metaWin.connect("notify::title", Lang.bind(this, this._on_title_change));
		this._ID_ws_changed = metaWin.connect("workspace-changed", Lang.bind(this, this._on_win_change_ws));
		
		if(metaWin.has_focus())
			this.set_focus();
	},
	
	new_icon: function(dock, ws_index) {
		let icon = new Icon(this, dock, this._app, ws_index, dock.settings.icon_label);
		this.icons.push(icon);
		
		if(Database.last_focus == this)
			icon.mark_focus(this);
		
		return icon;
	},
	index_icon: function(icon) {
		this.icons.push(icon);
	},
	remove_icon: function(i) {
		this.icons[i].destroy(false, false, this);
		this.icons.splice(i, 1);
	},
	
	set_all_ws: function(ws_index_to_skip) {
		//WARNING: asuming, that all_ws wasnt already true
		this.all_ws = true;
		let icons = this.icons,
			icons_i = icons.length,
			monitor = this.window.get_monitor();
		
		while(icons_i--) { //start from the back to make sure that new added icons are not processed as well - only theoretically possible
			icons[icons_i].dock.set_all_ws_win(this, ws_index_to_skip, monitor);
		}
		
		Database.index_all_ws_win(this);
	},
	unset_all_ws: function() {
		this.all_ws = false;
		let icons = this.icons,
			i = icons.length;
		
		while(i--) {
			if(icons[i].is_all_ws_clone) {
				this.remove_icon(i);
			}
		}
		Database.unindex_all_ws_win(this);
	},
	_on_win_change_ws: function(metaWin) {
		let ws = metaWin.get_workspace(),
			i = this.icons.length;
		
		if(!ws) //in gnome 3.16: when shell is reloaded
			return;
		let to = ws.index();
		while(i--) {//from the back because of splice!
			this.icons[i].dock.win_has_changed_ws(this, this.icons[i], to);
		}
	},
	moved_to_monitor: function(monitor) {
		let i = this.icons.length;
		
		while(i--) {
			this.icons[i].mark_monitor(monitor);
		}
	},
	
	_on_title_change: function() {
		let i = this.icons.length;
		
		while(i--) {
			this.icons[i].update_label();
		}
	},
	set_focus: function() {
		if(Database.last_focus)
			Database.last_focus.unset_focus();
		
		for(let k in this.icons) {
			this.icons[k].mark_focus(this);
		}
		Database.last_focus = this;
	},
	unset_focus: function() {
		for(let k in this.icons) {
			this.icons[k].unmark_focus();
		}
	},
	set_minimized: function() {
		for(let k in this.icons) {
			this.icons[k].mark_minimized();
		}
	},
	unset_minimized: function() {
		for(let k in this.icons) {
			this.icons[k].unmark_minimized();
		}
	},
	set_urgent: function() {
		let i = this.icons.length;
		
		while(i--) {
			this.icons[i].mark_urgent();
		}
	},
	
	destroy: function() {
		this.closed = true;
		this.window.disconnect(this._ID_unmanaged);
		this.window.disconnect(this._ID_focus);
		this.window.disconnect(this._ID_title);
		this.window.disconnect(this._ID_ws_changed);
		
		let i = this.icons.length;
		while(i--) { //there wont be splice anymore. So we wouldnt necesarily need while
			this.icons[i].destroy(false, true, this);
		}
		if(this.all_ws)
			Database.unindex_all_ws_win(this);
		this.icons = null;
	}
});
const Icon = new Lang.Class({
	Name: "Icon",

	winBox: null,
	app: null,
	dock: null,
	ws_index: 0, //not necessarily the REAL ws-index! Only where its saved internally
	app_windows: null, //array
	previous: false,
	next: false,

	actor: null,
	_box: null,
	_icon: null,
	_label: null,
	_hover_menu: null,
	_monitor_indicator: null,
	_ws_indicator: null,

	is_all_ws_clone: false,
	is_app_icon: false,
	is_fav_icon: false,
	is_focused: false,
	is_urgent: false,
	_drag_target: null, //for temp
	_drag_last_pos: 0,
	_insert_timeout: null,
	
	_init: function(winBox, dock, app, ws_index, icon_label) {
		let icon_size = dock.icon_size;
		this.winBox = winBox;
		this.dock = dock;
		this.app = app;
		this.ws_index = ws_index;
		
		
		this.actor = new St.Button(obj_size({ style_class: "icon_button",
				reactive: true,
				can_focus: true,
				x_align: St.Align.START,
				y_align: St.Align.START,
				width: icon_size, //width or height will be overwritten because of obj_size()
				height: icon_size //
			}, dock.vertical, icon_label ? icon_size + dock.label_length : icon_size));
		this.actor.set_pivot_point(0.5, 0.5);
		
		this._box = new St.BoxLayout({vertical: dock.vertical});
		this.actor.set_child(this._box);
		
		this._icon = app.create_icon_texture(icon_size-ICON_PADDING_SUM);
		this._icon.set_pivot_point(0.5, 0.5);
		this._box.add(this._icon);
		
		if(dock.settings.icon_num) {
			let num_box = new St.BoxLayout(obj_size({
					vertical: dock.vertical
				}, dock.vertical, 0));
			this._num_label = new St.Label(obj_translation({
					text: "1",
					width: 13,
					height: 13,
					style_class: "indicator_label"
				}, dock.vertical, -12));
			num_box.add(this._num_label);
			this._box.add(num_box, {x_align:St.Align.END, y_align:St.Align.END, x_fill:false, y_fill:false, x_expand: false, y_expand: false});
			this._num_label.hide();
		}
		
		if(icon_label)
			this.create_label();
		
		this.actor._delegate = this;
		this._icon._delegate = this;
		this._box._delegate = this;
		
		let dragable = DND.makeDraggable(this.actor);
		
		dragable.connect("drag-begin", Lang.bind(this, this._start_drag));
		dragable.connect("drag-end", Lang.bind(this, this._end_drag));
		
		this.actor.connect("button-release-event", Lang.bind(this, this._on_click));
		
		if(dock.settings.hover_menu)
			this.actor.connect("enter-event", Lang.bind(this, this._open_hover_menu));
	},
	create_label: function() {
		this._label = new St.Label({
				text: this.winBox.window.title,
				width: this.dock.label_length,
				style_class: "button_label"
			});
		this._label.set_pivot_point(0, 0.5);
		this._label._delegate = this;
		
		
		if(this.dock.intoPanel)
			this._box.add(this._label);
		else {
			if(this.dock.vertical) {
				//width is always expanded, and rotation isnt considered when calculating label-letters - so we need another box with no vertical:
				let box = new St.BoxLayout();
				box.set_pivot_point(0.5, 0.5);
				box.set_rotation_angle(Clutter.RotateAxis.Z_AXIS, 90);
				box.add(this._label);
				this._box.add(box);
			}
			else
				this._box.add(this._label, {y_align: St.Align.MIDDLE, y_fill: false});
		}
	},
	
	_open_hover_menu: function() {
		let dock = this.dock,
			metaWin = this.winBox.window,
			hover_box = dock.hover_box,
			wins = this.app.get_windows(),
			icon_size = dock.settings.icon_size,
			icon_label = dock.settings.icon_label,
			label_length = Math.max(HOVER_MENU_LENTGH, dock.label_length),
			currentWS = global.screen.get_active_workspace_index(),
			win, button, box;
		
		if(wins.length <= 1 || dock.menu.isOpen)
			return;
		
		hover_box.destroy_all_children();
		
		for(let k in wins) {
			win = wins[k];
			if(win.is_skip_taskbar())
				continue;
			button = new St.Button({ style_class: "icon_button",
				reactive: true,
				can_focus: true,
				width: label_length, //needed if vertical
				margin_left: Math.abs(win.get_workspace().index() - currentWS)*20,
				x_align: St.Align.START
			});
			box = new St.BoxLayout();
			box.add(this.app.create_icon_texture(icon_size-ICON_PADDING_SUM));
			button.set_child(box);
			
			
			box.add(new St.Label({
					text: win.title,
					style_class: "button_label"
				}), {y_align: St.Align.MIDDLE, y_fill: false});
			
			if(win.has_focus())
				button.add_style_class_name("focused");
			else if(win.minimized)
				button.opacity = 128;
			if(win == metaWin)
				button.add_style_class_name("selected");
				
			
			button.connect("button-release-event", Lang.bind(win, function() {
					toggle_win(this);
					dock._do_close_hover_menu();
				}));
			
			hover_box.add(button);
		}
		
		if(dock.vertical) {
			let monitor = LayoutManager.monitors[dock.settings.dock_monitor],
				max = monitor.y + monitor.height,
				pos = this.actor.get_transformed_position()[1],
				size = Math.max(wins.length * icon_size, this.dock.label_length + this.dock.icon_size);
			
			hover_box.margin_top = (size + pos > max) ? this.actor.get_y() - (size + pos - max) : this.actor.get_y();
			hover_box.set_height(size);
		}
		else {
			let monitor = LayoutManager.monitors[dock.settings.dock_monitor],
				max = monitor.x + monitor.width,
				pos = this.actor.get_transformed_position()[0],
				size = icon_size + label_length;
			
			hover_box.margin_left = (size + pos > max) ? this.actor.get_x() - (size + pos - max) : this.actor.get_x();
			//hover_box.margin_left = this.actor.get_x();
			hover_box.set_width(size);
		}
		dock.open_hover_menu();
	},
	
	update_label: function() {
		if(this._label) {
			if(this.is_app_icon && this.app_windows.length > 1)
				this._label.text = this.app.get_name();
			else
				this._label.text = this.winBox.window.title;
		}
	},
	
	_start_drag: function() {
		let pos = this.dock.vertical ? 1 : 0,
			dock_pos = this.dock.windowActor.get_transformed_position()[pos];
		Math.floor((this.actor.get_transformed_position()[pos]-this.dock_pos) / this.dock.settings.icon_size);
		Main.pushModal(this.actor);
		this.actor.opacity = 100;
		
		this._dragMonitor = {
				dragMotion: Lang.bind(this, this._on_dragMotion)
			};
		DND.addDragMonitor(this._dragMonitor);
		
		
		if(this.dock.menu.isOpen)
			this.dock.menu.close()
		if(this.dock.hover_menu_isOpen)
			this.dock.close_hover_menu();
	},
	_end_drag: function() {
		Main.popModal(this.actor);
		DND.removeDragMonitor(this._dragMonitor);
		this.actor.opacity = 255;
		
		let target = this._drag_target;
		if(!target || target == this)
			return;
			
			
		let after, before;
		if(this._drag_to_left) {
			after = "previous";
			before = "next";
		}
		else {
			after = "next";
			before = "previous";
		}
		
		//cut this. out
		if(this.next)
			this.next.previous = this.previous;
		if(this.previous)
			this.previous.next = this.next;
		else if(!this._drag_to_left)
			this.dock.windowList[this.ws_index] = this.next;
		
		//insert this.
		this[after] = target[after];
		this[before] = target;
		if(target[after])
			target[after][before] = this;
		else if(this._drag_to_left)
			this.dock.windowList[this.ws_index] = this;
		target[after] = this;
	},
	_on_dragMotion: function(dragEvent) {
		//WORKAROUND: there is a strange MetaWindowGroup only on the first monitor in foreground - but luckily its not reactive
		let dock = this.dock,
			target = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE, dragEvent.x, dragEvent.y);
			
		if(target._delegate && target._delegate.dock == this.dock) {
			if(target._delegate == this)
				return DND.DragMotionResult.MOVE_DROP;
				
			this._drag_target = target._delegate;
			
			if(dock.vertical) {
				var to_left = this._drag_last_pos > dragEvent.y;
				this._drag_last_pos = dragEvent.y;
			}
			else {
				var to_left = this._drag_last_pos > dragEvent.x;
				this._drag_last_pos = dragEvent.x;
			}
			dock.windowActor.remove_actor(this.actor);
			
			if(!to_left) { //if both are true or false - means they have the same value
				this._drag_to_left = false;
				dock.windowActor.insert_child_above(this.actor, target);
			}
			else {
				this._drag_to_left = true;
				dock.windowActor.insert_child_below(this.actor, target);
			}
			
			return DND.DragMotionResult.MOVE_DROP;
		}
		else {
			return DND.DragMotionResult.NO_DROP;
		}
	},
	
	getDragActor: function() { //for icon-dragging
		return new Clutter.Clone({ source: this.actor, opacity:180 });
	},
	getDragActorSource: function() { //for icon-dragging
		return this.actor;
	},
	handleDragOver: function(source, actor, eventX, eventY, time) { //for xdnd-drag
		this.do_map();
    },
	
	mark_focus: function(winBox) {
		this.is_focused = true;
		this.actor.add_style_class_name("focused");
		
		if(this.is_urgent)
			this.unmark_urgent();
		if(this.is_app_icon)
			this.winBox = winBox;
	},
	unmark_focus: function() {
		this.is_focused = false;
		//this.actor.remove_style_pseudo_class("focused");
		this.actor.remove_style_class_name("focused");
	},
	mark_minimized: function() {
		if(!this.is_app_icon)
			this._box.opacity = 128;
	},
	unmark_minimized: function() {
		if(!this.is_app_icon)
			this._box.opacity = 255;
	},
	mark_fav_icon: function() {
		this.is_fav_icon = true;
		
		this._icon.set_scale(0.7, 0.7);
		this._box.opacity = 128;
		
		if(!this.dock.settings.fav_label && this._label) {
			this._label.destroy();
			this._label = null;
			
			Tweener.addTween(this.actor, obj_size({
					time: 0.4,
					transition: "easeOutQuint"
				}, this.dock.vertical, this.dock.icon_size));
			
			--this.dock.iconLabel_count;
			
			this.dock.update_icon_size();
		}
		if(this._monitor_indicator) {
			this._monitor_indicator.destroy();
			this._monitor_indicator = null;
		}
		if(this._ws_indicator) {
			this._ws_indicator.destroy();
			this._ws_indicator = null;
		}
	},
	unmark_fav_icon: function(winBox) {
		winBox.index_icon(this);
		this.winBox = winBox;
		this.is_fav_icon = false;
		if(this.dock.settings.icon_label) {
			if(!this._label) {
				this.create_label();
				Tweener.addTween(this.actor, obj_size({
						time: 0.4,
						transition: "easeOutSine"
					}, this.dock.vertical, this.dock.icon_size + this.dock.label_length));
				
				++this.dock.iconLabel_count;
				this.dock.update_icon_size();
			}
			this.update_label();
		}
		
		this._icon.set_scale(1, 1);
		this._box.opacity = 255;
	},
	mark_urgent: function() {
		this.is_urgent = true;
		this.actor.add_style_class_name("urgent");
	},
	unmark_urgent: function() {
		this.is_urgent = false;
		this.actor.remove_style_class_name("urgent");
	},
	set_app_icon: function(app_name) {
		this.app_windows = [this.winBox];
		this.dock.applications[this.ws_index][app_name] = this;
		this.is_app_icon = true;
		this.update_label();
		if(this._monitor_indicator) {
			this._monitor_indicator.destroy();
			this._monitor_indicator = null;
		}
		if(this._ws_indicator) {
			this._ws_indicator.destroy();
			this._ws_indicator = null;
		}
	},
	
	create_arrow_indicator: function(side, translate) {
		let vertical = this.dock.vertical,
			vertical_box = new St.BoxLayout(obj_translation(obj_size({
				vertical: vertical,
				translation_x:translate,
				translation_y:translate
			}, vertical, 0), vertical, -12)),
			arrow = get_arrow(side);
		
		arrow.set_size(13, 13);
		arrow.set_style_class_name("indicator_label");
		arrow.set_x_expand(false);
		arrow.set_y_expand(false);
		vertical_box.add(arrow);
		
		this._box.add(vertical_box, {x_fill:false});//WORKAROUND: I cant find neither add() nor another way to set x_fill in the manual and x_fill is needed on vertical docklets
		this._box.set_child_above_sibling(vertical_box, this._icon); //if there is _label this has to be added before it 
		//this._box.insert_child_above(vertical_box, this._icon);
		
		return vertical_box;
	},
	mark_monitor: function(monitor) {
		if(!this.dock.settings.icon_monitor_indicator || (this.is_app_icon && this.app_windows.length > 1))
			return;
		
		let dock_monitor = this.dock.settings.dock_monitor;
	
		if(this._monitor_indicator)
			this._monitor_indicator.destroy();
		
		if(monitor == dock_monitor)
			return;
		
		this._monitor_indicator = this.create_arrow_indicator((monitor < dock_monitor) ? St.Side.LEFT : St.Side.RIGHT, -6);
	},
	mark_ws: function(ws_index) {
		if(!this.dock.settings.icon_ws_indicator || this.is_fav_icon || this.winBox.window.is_on_all_workspaces() || (this.is_app_icon && this.app_windows.length > 1))
			return;
		
		if(ws_index === undefined)
			ws_index = this.winBox.window.get_workspace().index();
		
		let current_ws = global.screen.get_active_workspace_index();
		
		if(this._ws_indicator)
			this._ws_indicator.destroy();
		if(current_ws == ws_index)
			return;
		
		this._ws_indicator = this.create_arrow_indicator((ws_index < current_ws) ? St.Side.TOP : St.Side.BOTTOM, 6);
	},
	
	remove: function() {
		this.dock.windowActor.remove_actor(this.actor);
	},
	insert: function(after) {
		Tweener.removeTweens(this.actor);
		let icon_size = this.dock.icon_size;
		
		this.actor.set_size(0, 0);
		this.actor.opacity = 0;
		
		if(after)
			this.dock.windowActor.insert_child_above(this.actor, after);
		else
			this.dock.windowActor.add(this.actor);
		
		Tweener.addTween(this.actor, obj_size({
			opacity: 255,
			width: icon_size,
			height: icon_size,
			time: 0.4,
			transition: "easeOutSine"
		}, this.dock.vertical, this._label ? this.dock.icon_size + this.dock.label_length : this.dock.icon_size));
	},
	insert_timeout: function(after) {
		this.actor.hide();
		if(after)
			this.dock.windowActor.insert_child_above(this.actor, after);
		else
			this.dock.windowActor.add(this.actor);
		
		this._insert_timeout = Mainloop.timeout_add(200, Lang.bind(this, function() {
				this.actor.show();
				this.actor.set_size(0, 0);
				this.actor.opacity = 0;
				let icon_size = this.dock.icon_size;
				Tweener.addTween(this.actor, obj_size({
					opacity: 255,
					width: icon_size,
					height: icon_size,
					time: 0.4,
					transition: "easeOutSine"
				}, this.dock.vertical, this._label ? this.dock.icon_size + this.dock.label_length : this.dock.icon_size));
				this._insert_timeout = false;
			}));
	},
	insert_no_size: function() {
		this.actor.opacity = 0;
		
		this.dock.windowActor.add(this.actor);
		Tweener.addTween(this.actor, {
			opacity: 255,
			time: 0.4,
			transition: "easeOutSine"
		});
	},
	increase_app_windows: function(winBox) {
		this.app_windows.push(winBox);
		
		if(this._monitor_indicator) {//now there is more than one app_windows
			this._monitor_indicator.destroy();
			this._monitor_indicator = null;
		}
		if(this._ws_indicator) {//now there is more than one app_windows
			this._ws_indicator.destroy();
			this._ws_indicator = null;
		}
		
		if(this.dock.settings.icon_num) {
			if(!this._num_label.visible)
				this._num_label.show();
			this._num_label.text = this.app_windows.length.toString();
		}
	},
	decrease_app_windows: function(winBox) {
		let i = this.app_windows.length;
		
		while(i--) {
			if(this.app_windows[i] == winBox) {
				this.app_windows.splice(i, 1);
				break;
			}
		}
		
		if(this.app_windows.length) {
			if(this.winBox == winBox)
				this.winBox = this.app_windows[0];
			
			this.mark_monitor(this.winBox.window.get_monitor());
			this.mark_ws(this.winBox.window.get_workspace().index());
			
			if(this._num_label) {
				if(this.app_windows.length == 1) {//we already checked if its less than 1
					this._num_label.hide();
					this.update_label();
				}
				else
					this._num_label.text = this.app_windows.length.toString();
			}
			return true;
		}
		else if(this._num_label) //is a fav-icon or will be destroyed
			this._num_label.hide();
		
		return false;
	},
	
	change_icon_size: function(size) {
		if(this._label)
			this._label.hide();
		this._icon.set_size(size-ICON_PADDING_SUM, size-ICON_PADDING_SUM);
		Tweener.addTween(this.actor, {
			width: size,
			height: size,
			time: 0.4,
			transition: "easeOutSine"
		});
	},
	change_label_size: function(size) {
		if(!this._label.visible)
			this._label.show();
		this._label.set_width(size);
		Tweener.addTween(this.actor, obj_size({
			time: 0.2,
			transition: "easeOutSine"
		}, this.dock.vertical, this.dock.icon_size + this.dock.label_length));
	},
	_on_click: function(actor, event) {
		let button = event.get_button();
		
		switch(button) {
			case 1://left
				if(this.winBox.window.has_focus())
					this.winBox.window.minimize();
				else {
					if(this.is_fav_icon) {
						this.app.open_new_window(this.ws_index);
					}
					else 
						this.do_map();
				}
			break;
			case 2://middle
				switch(this.dock.settings.middle_click_action) {
					case MIDDLE_CLICK.CLOSE:
						this.winBox.window.delete(global.get_current_time());
					break;
					case MIDDLE_CLICK.MINIMIZE:
						this.winBox.window.minimize();
					break;
					case MIDDLE_CLICK.NEW:
						this.app.open_new_window(global.screen.get_active_workspace_index());
					break;
					case MIDDLE_CLICK.QUIT:
						this.app.request_quit();
					break;
				}
			break;
			case 3://right
				this.dock.menu.open_for_icon(this);
		}
	},
	do_map: function() {
		//if(this.is_app_icon) {
			//let i = this.app_windows.length;
			//while(i--) {
				//this.app_windows[i].window.activate(global.get_current_time());
			//}
		//}
		//else if(!this.is_fav_icon)
			//this.winBox.window.activate(global.get_current_time());
		if(this.is_fav_icon)
			return
		if(this.winBox.window.get_workspace().index() != global.screen.get_active_workspace())
			this.winBox.window.get_workspace().activate(global.get_current_time());
		this.winBox.window.activate(global.get_current_time());
	},
	destroy: function(no_cancel, no_splice, winBox) { //FIXME: too many variables
		winBox = winBox || this.winBox; //app-icons need to know which window will be removed
		
		if(!no_splice) {//has to happen before no_cancel
			let icons = winBox.icons;
			for(let k in icons) {
				if(icons[k] == this) {
					icons.splice(k, 1);
					break;
				}
			}
		}
		
		if(!no_cancel) {
			if(this.is_app_icon) {
				if(this.decrease_app_windows(winBox))//if there are still windows out there
					return true; //true: icon hasnt been changed
				delete this.dock.applications[this.ws_index][this.app.get_id()];
			}
			
			let fav = this.dock.fav_icons[this.ws_index],
				app_name = this.app.get_id();
			
			if(this.dock.settings.favorites && fav.hasOwnProperty(app_name)) {
				let other_icon = this.dock.get_first_app_window(this.dock.windowList[this.ws_index], this.app);
				if(other_icon == this)//its very likely that .this will be found - just search again
					other_icon = this.dock.get_first_app_window(this.next, this.app);
				
				if(other_icon.app != this.app) {
				
					this.mark_fav_icon();
					fav[app_name] = this;
					if(this.is_focused)
						this.unmark_focus();
					if(this.is_urgent)
						this.unmark_urgent();
					
					let obj = Object.create(FAV_WIN_MASC);
					obj.window.title = this.app.get_name();
					this.winBox = obj;
					this.update_label();
					
					return false;
				}
			}
		}
		
		this.dock.remove_icon(this);
		
		
		
		//destroy actor:
		//
		if(this._insert_timeout)
			Mainloop.source_remove(this._insert_timeout);
		Tweener.addTween(this._icon, {
			scale_x: 2,
			scale_y: 2,
			time: 0.3,
			transition: "easeInOutSine"
		});
		if(this._label)
			Tweener.addTween(this._label, {
				scale_x: 2,
				scale_y: 2,
				time: 0.3,
				transition: "easeInOutSine"
			});
		
		Tweener.addTween(this.actor, {
			opacity: 0,
			width: 0,
			height:0,
			time: 0.4,
			transition: "easeOutSine",
			onComplete: Lang.bind(this.actor, function() {
				this.destroy();
			})
		});
		return false;
	}
});

const PanelDocklet_dbus = {
	_dbus: null,
	_docklets: null, //array
	_current_selected: null,
	
    init: function(docklets) {
		this._docklets = docklets;
        this._dbus = Gio.DBusExportedObject.wrapJSObject(PanelDocklet_iface, this);
        this._dbus.export(Gio.DBus.session, "/org/gnome/Shell/Extensions/panelDocklet");
    },

	get_monitors: function() {
		let monitors = LayoutManager.monitors,
			i = monitors.length,
			out = [],
			m;
		
		while(i--) {
			m = monitors[i];
			out[i] = [m.x, m.y, m.width, m.height];
		}
		return GLib.Variant.new("(aai)", [out]);
	},
	
    createDock: function() {
		Settings.create_cache();
		
		let i = this._docklets.length;
		this._docklets[i] = new Docklet(i);
		this.selectDocklet(i);
    },
    updateDock: function(settings, key) {
		if(!this._current_selected) {
			if(key == "dock-count") {
				if(Settings.get_int("dock-count") > this._docklets.length)
					this.createDock();
			}
			return;
		}
		let i = this._current_selected.id,
			docklets = this._docklets;
		Settings.create_cache();
		
		this.deselectDocklet();
		docklets[i].destroy(true);
		docklets[i] = new Docklet(i); //cant use d
		this.selectDocklet(i);
		
		solve_shaking();
	},
    deleteDock: function(i) {
		this.deselectDocklet();
		this._docklets[i].destroy();
		this._docklets.splice(i, 1);
		Settings.cache.splice(i, 1);
		//Settings.create_cache();
	},
	selectDocklet: function(i) {
		if(this._current_selected)
			this._current_selected.deselect();
		
		if(i < this._docklets.length)//can happen when a new docklet is created
			(this._current_selected = this._docklets[i]).select();
	},
	deselectDocklet: function(i) {
		if(this._current_selected) {
			this._current_selected.deselect();
			this._current_selected = null;
		}
	},
	
	
    destroy: function() {
		this._dbus.unexport();
		this._docklets = null;
	}
};

function obj_size(o, vertical, i) {
	if(vertical)
		o.height = i;
	else
		o.width = i;
	return o;
}
function obj_translation(o, vertical, i) {
	if(vertical)
		o.translation_y = i;
	else
		o.translation_x = i;
	return o;
}
function toggle_win(win) {
	if(win.has_focus())
		win.minimize();
	else {
		let ws = win.get_workspace();
		if(ws != global.screen.get_active_workspace())
			ws.activate(global.get_current_time());
		
		win.activate(global.get_current_time());
	}
}
function get_app(metaWin) {
	return Tracker.get_window_app(metaWin);
}

function solve_shaking() {
	//WORKAROUND: size-tweens create shaking on other actors if they have no fixed size
	
	let docklets = PanelDocklet_dbus._docklets,
		i = docklets.length,
		panel_left = false,
		panel_middle = false,
		panel_middle_right = false,
		panel_right = false,
		item, dock;
	
	while(i--) {
		if(!docklets[i].intoPanel)
			continue;
		dock = docklets[i];
		switch(dock.settings.dock_orientation) {
			case DOCK_POS.PANEL_LEFT:
				panel_left = true;
				if(!dock.settings.dock_fixed_size) {
					dock.icon_space -= Main.panel.statusArea.activities.actor.get_width() + Main.panel.statusArea.dateMenu.actor.get_width()/2;
					dock.update_icon_size();
				}
			break;
			case DOCK_POS.PANEL_MIDDLE_LEFT:
			case DOCK_POS.PANEL_MIDDLE_RIGHT:
				panel_middle = true;
				log(Main.panel._rightBox.get_width())
				if(!dock.settings.dock_fixed_size) {
					dock.icon_space -= Main.panel._rightBox.get_width() + Main.panel.statusArea.activities.actor.get_width() + Main.panel.statusArea.dateMenu.actor.get_width();
					dock.update_icon_size();
				}
			break;
			case DOCK_POS.PANEL_RIGHT:
				panel_right = true;
				if(!dock.settings.dock_fixed_size) {
					dock.icon_space -= Main.panel.statusArea.dateMenu.actor.get_width() + (Main.panel._rightBox.get_width() - dock.actor.get_width());
					dock.update_icon_size();
				}
			break;
		}
	}
	if(panel_left) {
		item = Main.panel.statusArea.activities.actor;
		item.set_width(item.get_width());
	}
	
	if(panel_middle) {
		item = Main.panel.statusArea.dateMenu.actor;
		item.set_width(item.get_width());
	}
		
	//if(panel_right) {
		//item = Main.panel._rightBox;
		//item.set_width(item.get_width());
	//}
}

function init(metadata) {
}
function enable() {
	Settings = Convenience.getSettings();
	Settings.create_cache();
	Convenience.initTranslations();
	Database.init();

	let docklets = [],
		i = parseInt(Settings.get_int("dock-count"));
	while(i--) {
		docklets[i] = new Docklet(i);
	}
	
	PanelDocklet_dbus.init(docklets);
	
	if(Settings.get_boolean("first-use")) {
		Main.Util.trySpawnCommandLine("gnome-shell-extension-prefs " + Extension.metadata.uuid);
		Settings.set_boolean("first-use", false);
	}
	//Settings.connect("change-event", Lang.bind(PanelDocklet_dbus, PanelDocklet_dbus.updateDock)) //crashes shell for some reason
	Settings.connect("changed", Lang.bind(PanelDocklet_dbus, PanelDocklet_dbus.updateDock));
	
	Mainloop.timeout_add(2000, solve_shaking); //wait until other extension had a chance to do their changes
}
function disable() {
	for(let k in PanelDocklet_dbus._docklets) {//FIXME: not very pretty
		PanelDocklet_dbus._docklets[k].destroy();
	}
	PanelDocklet_dbus.destroy();
	Database.destroy();
	
	//remove shaking-workaround:
	let cache = Settings.cache,
		i = Settings.cache.length;
	while(i--) {
		switch(cache[i].dock_orientation) {
			case DOCK_POS.PANEL_LEFT:
				Main.panel.statusArea.activities.actor.set_width(-1);
			break;
			case DOCK_POS.PANEL_MIDDLE_LEFT:
			case DOCK_POS.PANEL_MIDDLE_RIGHT:
				Main.panel.statusArea.dateMenu.actor.set_width(-1);
			//case DOCK_POS.PANEL_RIGHT:
				//item = Main.panel._rightBox.set_width(-1);
		}
	}
}
